{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE TupleSections #-}
{-# LANGUAGE Rank2Types #-}

module Imperative (Applicative(..), liftM, ap, MonadTrans(..), makeRS) where

import qualified Data.Char
import Control.Applicative (Applicative(..))
import Control.Monad (liftM, liftM2, ap)
import Control.Monad.Trans.Class (MonadTrans(..))
import Language.Haskell.TH

class (MonadTrans t) => MonadTrans1 t where
  lift1 :: (Monad m1, Monad m2) => (forall a. m1 a -> m2 a) -> t m1 b -> t m2 b

functorApplicative :: CxtQ -> TypeQ -> DecsQ
functorApplicative cxt monad =
  sequence [
    instanceD cxt [t|Functor $(monad)|] [
      funD (mkName "fmap") [clause [] (normalB [|liftM|]) []]
    ],
    instanceD cxt [t|Applicative $(monad)|] [
      funD (mkName "pure") [clause [] (normalB [|return|]) []],
      funD (mkName "<*>") [clause [] (normalB [|ap|]) []]
    ]
  ]

makeRS :: String -> TypeQ -> DecsQ
makeRS entity qtp = if not ('a' <= head entity && head entity <= 'z')
  then fail "first symbol of a entity is not \"a\"..\"z\""
  else do {
    let {
      upper = [Data.Char.toUpper (head entity)] ++ tail entity;

      nentity = mkName entity;
      nsetEntity = mkName $ "set" ++ upper;
      nEntityR = mkName $ upper ++ "R";
      nEntityS = mkName $ upper ++ "S";
      nEntityReaderT = mkName $ upper ++ "ReaderT";
      nEntityStateT = mkName $ upper ++ "StateT";
      nrunEntityReaderT = mkName $ "run" ++ upper ++ "ReaderT";
      nrunEntityStateT = mkName $ "run" ++ upper ++ "StateT";
      nRemoveEntityReaderT = mkName $ "Remove" ++ upper ++ "ReaderT";
      nsnapshotEntity = mkName $ "snapshot" ++ upper;

      qentity = varE nentity;
      qsetEntity = varE nsetEntity;
      qEntityR = conT nEntityR;
      qEntityS = conT nEntityS;
      qTypeEntityReaderT = conT nEntityReaderT;
      qCtorEntityReaderT = conE nEntityReaderT;
      qTypeEntityStateT = conT nEntityStateT;
      qCtorEntityStateT = conE nEntityStateT;
      qrunEntityReaderT = varE nrunEntityReaderT;
      qrunEntityStateT = varE nrunEntityStateT;
      qRemoveEntityReaderT = varT nRemoveEntityReaderT;
      qsnapshotEntity = varE nsnapshotEntity;
    };

    nm <- newName "m";
    nt <- newName "t";
    na <- newName "a";
    nm2 <- newName "m2";

    let {
      qm = varT nm;
      qt = varT nt;
      qa = varT na;
      qm2 = varT nm2;

      app = liftM2 (++);
    };

    -- EntityR
    sequence [
      classD (sequence [classP (mkName "Monad") [qm]]) nEntityR [PlainTV nm] [] [
        sigD nentity (appT qm qtp)
      ]
    ] `app`
    sequence [
      instanceD (sequence [classP (mkName "MonadTrans") [qt], classP nEntityR [qm], classP (mkName "Monad") [appT qt qm]]) (appT qEntityR (appT qt qm)) [
        funD nentity [clause [] (normalB $ [|lift $(qentity)|]) []]
      ]
    ] `app`

    -- EntityReaderT
    sequence [
      newtypeD (sequence []) nEntityReaderT [PlainTV nm, PlainTV na] (recC nEntityReaderT [liftM2 (nrunEntityReaderT,,) notStrict [t|$(qtp) -> $(qm) $(qa)|]]) []
    ] `app`
    [d|
      instance MonadTrans $(qTypeEntityReaderT) where {
        lift a = $(qCtorEntityReaderT) (\_ -> a);
      }
    |] `app`
    [d|
      instance (Monad m) => Monad ($(qTypeEntityReaderT) m) where {
        a >>= b = $(qCtorEntityReaderT) (\r -> do {
          val <- $(qrunEntityReaderT) a r;
          $(qrunEntityReaderT) (b val) r;
        });
        return = lift . return;
      }
    |] `app`
    functorApplicative (sequence [classP (mkName "Monad") [qm]]) (appT qTypeEntityReaderT qm) `app`
    sequence [
      instanceD (sequence [classP (mkName "Monad") [qm]]) (appT qEntityR (appT qTypeEntityReaderT qm)) [
        funD nentity [clause [] (normalB $ [|$(qCtorEntityReaderT) return|]) []]
      ]
    ] `app`
    [d|
      instance MonadTrans1 $(qTypeEntityReaderT) where {
        lift1 f m = $(qCtorEntityReaderT) (\r -> f ($(qrunEntityReaderT) m r));
      }
    |] `app`

    -- EntityS
    sequence [
      classD (sequence [classP nEntityR [qm]]) nEntityS [PlainTV nm] [] [
        sigD nsetEntity [t|$(qtp) -> $(qm) ()|]
      ]
    ] `app`
    sequence [
      instanceD (sequence [classP (mkName "MonadTrans") [qt], classP nEntityS [qm], classP (mkName "Monad") [appT qt qm]]) (appT qEntityS (appT qt qm)) [
        funD nsetEntity [clause [] (normalB $ [|\s -> lift ($(qsetEntity) s)|]) []]
      ]
    ] `app`

    -- EntityStateT
    sequence [
      newtypeD (sequence []) nEntityStateT [PlainTV nm, PlainTV na] (recC nEntityStateT [liftM2 (nrunEntityStateT,,) notStrict [t|$(qtp) -> $(qm) ($(qa), $(qtp))|]]) []
    ] `app`
    [d|
      instance MonadTrans $(qTypeEntityStateT) where {
        lift a = $(qCtorEntityStateT) (\s -> do {
          unboxed <- a;
          return (unboxed, s);
        });
      }
    |] `app`
    [d|
      instance (Monad m) => Monad ($(qTypeEntityStateT) m) where {
        a >>= b = $(qCtorEntityStateT) (\s1 -> do {
          (val, s2) <- $(qrunEntityStateT) a s1;
          $(qrunEntityStateT) (b val) s2;
        });
        return = lift . return;
      }
    |] `app`
    functorApplicative (sequence [classP (mkName "Monad") [qm]]) (appT qTypeEntityStateT qm) `app`
    sequence [
      instanceD (sequence [classP (mkName "Monad") [qm]]) (appT qEntityR (appT qTypeEntityStateT qm)) [
        funD nentity [clause [] (normalB $ [|$(qCtorEntityStateT) (\s -> return (s, s))|]) []]
      ]
    ] `app`
    sequence [
      instanceD (sequence [classP (mkName "Monad") [qm]]) (appT qEntityS (appT qTypeEntityStateT qm)) [
        funD nsetEntity [clause [] (normalB $ [|\s -> $(qCtorEntityStateT) (\_ -> return ((), s))|]) []]
      ]
    ] `app`
    [d|
      instance MonadTrans1 $(qTypeEntityStateT) where {
        lift1 f m = $(qCtorEntityStateT) (\s -> f ($(qrunEntityStateT) m s));
      }
    |] `app`

    -- rename RemoveEntityReaderT
    -- RemoveEntityReaderT
    sequence [
      --classD (sequence [classP (mkName "Monad") [qm], classP (mkName "Monad") [qm2]]) nRemoveEntityReaderT [PlainTV nm, PlainTV nm2] [FunDep [nm] [nm2]] [
        --sigD nsnapshotEntity [t|$(qTypeEntityStateT) $(qm2) a -> $(qm) a|]
      --]
    ] `app`
    --sequence [
      --instanceD (sequence [...]) (appT (appT qRemoveEntityReaderT (appT qreader
      sequence [];
  }
