-- TODO: what to export?
module Kernel where

import IAmAFreeVariable

sortOfSort :: Sort -> Maybe Sort
sortOfSort Metaprop = Just MetapropSort
sortOfSort Type     = Just TypeSort
sortOfSort _        = Nothing

-- TODO: всё дописать
rule :: Sort -> Sort -> Either String Sort
rule a b = throwError $ "internal error: rule: no rule (" ++ show a ++ ", " ++ show b ++ ", _)"

merge :: [Maybe (String, PTerm)] -> [Maybe (String, PTerm)] -> Either String [Maybe (String, PTerm)]

merge [] ys = return ys

merge xs [] = return xs

merge (Nothing:xs) (y:ys) = do {
  result <- merge xs ys;
  return $ y:result;
}

merge (x:xs) (Nothing:ys) = do {
  result <- merge xs ys;
  return $ x:result;
}

merge ((Just x):xs) ((Just y):ys) = do {
  if x /= y
    then throwError "internal error: merge: contexts differ"
    else return ();
  result <- merge xs ys;
  return $ (Just x):result;
}

alphaEq :: Int -> PTerm -> PTerm -> Bool
alphaEq n (PSort a1                  ) (PSort a2                  ) = a1 == a2
alphaEq n (PFree a1                  ) (PFree a2                  ) = a1 == a2
alphaEq n (PApplication a1 _ _ b1 _ _) (PApplication a2 _ _ b2 _ _) = alphaEq a1 a2 && alphaEq b1 b2
alphaEq n (PLambda _ a1 _ b1 _ _     ) (PLambda _ a2 _ b2 _ _     ) = alphaEq a1 a2 && alphaEq (n + 1) (instantiate (PFree n) b1) (instantiate (PFree n) b2)
alphaEq n (PPi _ a1 _ b1 _           ) (PPi _ a2 _ b2 _           ) = alphaEq a1 a2 && alphaEq (n + 1) (instantiate (PFree n) b1) (instantiate (PFree n) b2)
alphaEq n _ _ = False

betaNormal :: Int -> PTerm -> PTerm
betaNormal n (PSort a) = PSort a
betaNormal n (PFree a) = PFree a
betaNormal n (PApplication a at att b bt btt) = let {
  a2 = betaNormal n a;
  b2 = betaNormal n b;
} in case a2 of {
  PLambda _ _ _ scope _ _ -> betaNormal n $ instantiate b2 scope;
  _ -> PApplication a2 at att b2 bt btt;
}
betaNormal n (PLambda x a at b bt btt) = PLambda x (betaNormal a) at (abstract n $ betaNormal (n + 1) $ instantiate (PFree n) b) bt btt
betaNormal n (PPi x a at b bt) = PPi x (betaNormal a) at (abstract n $ betaNormal (n + 1) $ instantiate (PFree n) b) bt

alphaBetaEq :: Int -> PTerm -> PTerm -> Bool
alphaBetaEq n a b = alphaEq n (betaNormal n a) (betaNormal n b)

abstractInternal :: [Maybe (String, PTerm)] -> PTerm -> PTerm -> [Maybe (String, PTerm)] -> [PTerm] -> Either String ([Maybe (String, PTerm)], String, [Scope])
abstractInternal ac a at bc b = do {
  if a /= PFree (length ac - 1)
    then throwError "internal error: abstractInternal: a /= PFree (length ac - 1)"
    else return ();
  case last ac of {
    Nothing -> throwError "internal error: abstractInternal: last ac == Nothing";
    Just (x, at2) -> do {
      if at /= at2
        then throwError "internal error: abstractInternal: at /= at2"
        else return ();
      if length ac < length bc
        then throwError "internal error: abstractInternal: length ac < length bc"
        else return ();
      c <- merge ac bc;
      return (init c, x, map (abstract (length ac - 1)) b);
    };
  };
}

-- END: может быть, перевернуть список так, чтобы новое было вначале, для ускорения. То же для остальных списков
-- TODO: Statement содержит Maybe для trace'а того, что использовалось. Реализовать эту фичу. Возможно, можно не хранить в Statement контекст, а использовать вместо этого некую монаду
data Statement = Statement [Maybe (String, PTerm)] PTerm PTerm (Maybe Sort)

stMetaprop :: Statement
stMetaprop = Statement [] (PSort Metaprop) (PSort MetapropSort) Nothing

stType :: Statement
stType = Statement [] (PSort Type) (PSort TypeSort) Nothing

start :: Int -> String -> Statement -> Either String Statement
start n x (Statement c a at _) = do {
  if n < length c
    then throwError "internal error: start: n < length c"
    else return ();
  case at of {
    PSort s -> return $ Statement (c ++ replicate (n - length c) Nothing ++ [Just (x, a)]) (PFree n) a (Just s);
    -- TODO: beta
    _ -> throwError "internal error: start: at /= PSort _";
  };
}

arrow :: Statement -> Statement -> Either String Statement
arrow (Statement ac a at att) (Statement bc b bt _) = do {
  (c, x, [bScope]) <- abstractInternal ac a at bc [b];
  case att of {
    Nothing -> throwError "internal error: arrow: att == Nothing";
    Just att2 -> case bt of {
      -- TODO: beta
      PSort bt2 -> do {
        resultT <- rule att2 bt2;
        return $ Statement c (Pi x at att2 bScope bt2) (PSort resultT) (sortOfSort resultT);
      };
      _ -> throwError "internal error: arrow: bt /= PSort _";
    };
  };
}

application :: Statement -> Statement -> Either String Statement
application (Statement ac a at att) (Statement bc b bt btt) = do {
  c <- merge ac bc;
  case at of { -- TODO: beta
    PPi _ bt2 _ resultT resultTT -> do {
      if not $ alphaEq (length c) bt2 bt -- TODO: beta!
        then throwError "internal error: application: type mismatch"
        else return ();
      return $ Statement c (PApplication a at att b bt btt) (instantiate b resultT) (Just resultTT);
    };
    _ -> throwError "internal error: application: application of non-function";
  };
}

abstraction :: Statement -> Statement -> Either String Statement
abstraction (Statement ac a at att) (Statement bc b bt btt) = do {
  (c, x, [bScope, btScope]) <- abstractInternal ac a at bc [b, bt];
  case att of {
    Nothing -> throwError "internal error: abstraction: att == Nothing";
    Just att2 -> case btt of {
      Nothing -> throwError "internal error: abstraction: btt == Nothing";
      Just btt2 -> do {
        resultTT <- rule att2 btt2;
        return $ Statement c (PLambda x at att2 bScope btScope btt2) (PPi x at att2 btScope btt2) (Just resultTT);
      };
    };
  };
}

hasSort :: Statement -> Sort -> Bool
hasSort (Statement _ _ at _) s = at == PSort s -- TODO: beta?

hasSortOfPType :: Statement -> Sort -> Bool
hasSortOfPType (Statement _ _ _ att) s = att == Just s

pTypeOf :: Statement -> Either String Statement
pTypeOf (Statement c _ at att) = case att of {
  Nothing -> throwError "internal error: pTypeOf: no PTS type";
  Just att2 -> return $ Statement c at (PSort att2) (sortOfSort att2);
}

stEq :: Statement -> Statement -> Either String Bool
stEq (Statement ac a _ _) (Statement bc b _ _) = do {
  c <- merge ac bc;
  return $ alphaBetaEq (length c) a b;
}

-- TODO: допилить для всех применений
-- TODO: нужен ли вот этот трюк с подрезанием контекста у сортов и Free?
stCase :: Statement -> (Statement -> a) -> (Int -> Statement -> a) -> (Statement -> Statement -> a) -> (Int -> Statement -> Statement -> a) -> (Int -> Statement -> Statement -> a) -> a
stCase (Statement c (PSort a                       ) at att) f _ _ _ _ = f (Statement [] (PSort a) at att)
stCase (Statement c (PFree a                       ) at att) _ f _ _ _ = f a (Statement (take (a + 1) c) (PFree a) at att)
stCase (Statement c (PApplication a at att b bt btt) _  _  ) _ _ f _ _ = f (Statement c a at (Just att)) (Statement c b bt (Just btt))
stCase (Statement c (PLambda x a at b bt btt       ) _  _  ) _ _ _ f _ = f (length c) (Statement c a (PSort at) (sortOfSort at)) (Statement (c ++ [Just (x, a)]) (instantiate (PFree $ length c) b) (instantiate (PFree $ length c) bt) (Just btt))
stCase (Statement c (PPi x a at b bt               ) _  _  ) _ _ _ _ f = f (length c) (Statement c a (PSort at) (sortOfSort at)) (Statement (c ++ [Just (x, a)]) (instantiate (PFree $ length c) b) (PSort bt) sortOfSort bt)
