module Etc where

import qualified Data.Map
import Text.Parsec

import Kernel
import Parser

emptyEtcState = Nothing

>>>>>etcDecl :: Parsec String OuterState ()
etcDecl = do {
            st <- getState;
            oReserved "setup_deduction";
            impl <- oIdentifier;   -- A --> B
            mp <- oIdentifier;     -- [|A --> B; A|] ==> B
            aImplA <- oIdentifier; -- A --> A
            a1 <- oIdentifier;     -- A --> (B --> A)
            a2 <- oIdentifier;     -- (A --> (B --> C)) --> ((A --> B) --> (A --> C))
            setState st { stEtc = Just (impl, mp, aImplA, a1, a2) };
            return ();
          }

-- TODO: нужно проверять шаги на предмет непоявления новых переменных и типовых переменных

data ProofExpr = ProofExpr String [ProofExpr]

proofExpr :: Parsec String OuterState ProofExpr
proofExpr =   do {
                i <- oIdentifier;
                ps <- option [] $ between (oReservedOp "(") (oReservedOp ")") $ sepBy1 proofExpr (oReservedOp ",");
                return (ProofExpr i ps);
              }

data Step = Have String Expr Reason | Assume String Expr [Step]
data Reason = ByRule ProofExpr | Qed [Step] Reason

step :: Parsec String OuterState Step
step  =   try (do {
            oReserved "have";
            i <- oIdentifier;
            oReservedOp ":";
            e <- oExpr;
            p <- proof;
            return (Have i e p);
          })
      <|> do {
            oReserved "have";
            e <- oExpr;
            p <- proof;
            return (Have "this" e p);
          }
      <|> do {
            oReservedOp "{";
            oReserved "assume";
            i <- option "this" $ do {
              i <- oIdentifier;
              oReservedOp ":";
              return i;
            };
            assm <- oExpr;
            steps <- many step;
            oReservedOp "}";
            return $ Assume i assm steps;
          }

proof :: Parsec String OuterState Reason
proof =   do {
            oReserved "by";
            e <- proofExpr;
            return (ByRule e);
          }
      <|> do {
            oReserved "proof";
            steps <- many step;
            oReserved "show";
            p <- proof;
            oReserved "qed";
            return (Qed steps p);
          }

checkExpr :: OuterState -> Data.Map.Map String Proof -> ProofExpr -> Proof
checkExpr ctx m (ProofExpr i ps) = let {
  thm = case Data.Map.lookup i (stTheorems ctx) of {
    Just thm -> thm;
    Nothing -> case Data.Map.lookup i m of {
      Just fact -> if null ps then fact else error (i ++ ": argumented fact");
      Nothing -> error $ i ++ ": theorem or fact not found";
    };
  };
} in
  foldl (\r (i, p) -> proofCut (checkExpr ctx m p) i r) thm (zip (map (\x -> "$" ++ show x) [1..]) ps)

>>>>deduction :: OuterState -> Expr -> Proof -> Proof
deduction ctx assm proof = case stEtc ctx of {
  Nothing -> error "deduction denied";
  Just (implN, mpN, aImplAN, a1N, a2N) -> case map (\x -> Data.Map.lookup x $ stTheorems ctx) [mpN, aImplAN, a1N, a2N] of {
    [Just mpP, Just aImplAP, Just a1P, Just a2P] -> let {
      im a b = exprApply (exprApply (exprIdent (stConsts $ stInner ctx) implN) a) b;
      mp a b = proofCut b "$2" $ proofCut a "$1" mpP;
    } in proofCase proof
      (\assm2 -> if exprEq assm2 assm
        then proofTyped (im assm assm) aImplAP
        else mp (proofTyped (im assm2 (im assm assm2)) a1P) (proofAssume assm2))
      (\e rule premises -> if rule == mpN
        then case premises of {
          [bc, c] -> mp (mp a2P (deduction ctx assm bc)) (deduction ctx assm c);
          _ -> error "deduction: MP has premises count not equal to 2";
        }
        else if null premises
          then case Data.Map.lookup rule $ stTheorems ctx of {
            Nothing -> error "internal error";
            Just thm -> mp (proofTyped (im e (im assm e)) a1P) (proofTyped e thm);
          }
          else error ("deduction: " ++ rule ++ ": the rule denied in deduction"));
    _ -> error "deduction: some theorems for deduction not found";
  };
}

checkStep :: OuterState -> Step -> Data.Map.Map String Proof -> Data.Map.Map String Proof

checkStep ctx (Have i e p) m = let obj = checkReason ctx m e p in (Data.Map.insert i obj $ Data.Map.insert "this" obj $ m)

checkStep ctx (Assume assmIdent assm steps) m = let assmObj = proofAssume assm in
  case Data.Map.lookup "this" $ checkSteps ctx steps $ Data.Map.insert assmIdent assmObj $ Data.Map.insert "assm" assmObj $ Data.Map.insert "this" assmObj m of {
    Nothing -> error "internal error";
    Just deduct -> Data.Map.insert "this" (deduction ctx assm deduct) m;
  }

checkSteps :: OuterState -> [Step] -> Data.Map.Map String Proof -> Data.Map.Map String Proof
checkSteps ctx steps m = foldl (\m step -> checkStep ctx step m) m steps

checkReason :: OuterState  -> Data.Map.Map String Proof -> Expr -> Reason -> Proof
checkReason ctx m e (ByRule p) = proofTyped e (checkExpr ctx m p)
checkReason ctx m e (Qed steps sh) = checkReason ctx (checkSteps ctx steps m) e sh
