-- TODO Doc: возможно, parsec позволил бы обойтись без деления на inner and outer languages, но я оставил разделение для возможности смены языка реализации и для облегчения альтернативных реализаций. По той же причине запрещены миксфиксы разных видов одного приоритета

{-# LANGUAGE DeriveGeneric #-}

module Parser where

import qualified GHC.Generics
import qualified Data.Set
import qualified Data.Map
import qualified Control.DeepSeq
import Text.Parsec
import qualified Text.Parsec.Error
import qualified Text.Parsec.Token
import qualified Text.Parsec.Language
import qualified Text.Parsec.Expr

import Kernel

data Mixfixes = Prefix [(String, String)] | Infix [(String, String)] | Infixl [(String, String)] | Infixr [(String, String)] deriving (GHC.Generics.Generic)
instance Control.DeepSeq.NFData Mixfixes

-- TODO: вынести
type EtcState = Maybe (String, String, String, String, String)

data InnerState = InnerState { stSimpleTypes :: Data.Set.Set String, stConsts :: Data.Map.Map String Type, stConstMixfixes :: Data.Set.Set (String, String), stBinders :: Data.Set.Set (String, String), stMixfixes :: Data.Map.Map Int Mixfixes } deriving (GHC.Generics.Generic)
instance Control.DeepSeq.NFData InnerState

data OuterState = OuterState { stInner :: InnerState, stTheorems :: Data.Map.Map String Proof, stEtc :: EtcState } deriving (GHC.Generics.Generic)
instance Control.DeepSeq.NFData OuterState

innerLangDef :: Text.Parsec.Token.LanguageDef InnerState
innerLangDef = Text.Parsec.Language.emptyDef
  { Text.Parsec.Token.commentStart    = ""
  , Text.Parsec.Token.commentEnd      = ""
  , Text.Parsec.Token.commentLine     = ""
  , Text.Parsec.Token.nestedComments  = False
  , Text.Parsec.Token.identStart      = letter <|> oneOf "_'"
  , Text.Parsec.Token.identLetter     = alphaNum <|> oneOf "_'"
  , Text.Parsec.Token.opStart         = oneOf ""
  , Text.Parsec.Token.opLetter        = oneOf ""
  , Text.Parsec.Token.reservedOpNames = ["=>", "::", "%", "."]
  , Text.Parsec.Token.reservedNames   = []
  , Text.Parsec.Token.caseSensitive   = True
  }

outerLangDef :: Text.Parsec.Token.LanguageDef OuterState
outerLangDef = Text.Parsec.Language.emptyDef
  { Text.Parsec.Token.commentStart    = "(*"
  , Text.Parsec.Token.commentEnd      = "*)"
  , Text.Parsec.Token.commentLine     = ""
  , Text.Parsec.Token.nestedComments  = False
  , Text.Parsec.Token.identStart      = letter <|> oneOf "_'"
  , Text.Parsec.Token.identLetter     = alphaNum <|> oneOf "_'"
  , Text.Parsec.Token.opStart         = oneOf ""
  , Text.Parsec.Token.opLetter        = oneOf ""
  , Text.Parsec.Token.reservedOpNames = ["::", "==>", "[|", ";", "|]", ":", "(", ",", ")", "{", "}"]
  , Text.Parsec.Token.reservedNames   = ["consts", "axioms", "have", "by", "proof", "show", "qed", "lemma", "theorem", "assumes", "shows", "prefix", "infix", "infixl", "infixr", "binder", "setup_deduction", "assume"]
  , Text.Parsec.Token.caseSensitive   = True
  }

innerLexer :: Text.Parsec.Token.TokenParser InnerState
innerLexer = Text.Parsec.Token.makeTokenParser innerLangDef

outerLexer :: Text.Parsec.Token.TokenParser OuterState
outerLexer = Text.Parsec.Token.makeTokenParser outerLangDef

iIdentifier = Text.Parsec.Token.identifier innerLexer
iParens     = Text.Parsec.Token.parens     innerLexer
iReservedOp = Text.Parsec.Token.reservedOp innerLexer
iWhiteSpace = Text.Parsec.Token.whiteSpace innerLexer

oIdentifier = Text.Parsec.Token.identifier outerLexer
oReserved   = Text.Parsec.Token.reserved   outerLexer
oReservedOp = Text.Parsec.Token.reservedOp outerLexer
oWhiteSpace = Text.Parsec.Token.whiteSpace outerLexer
oDecimal    = Text.Parsec.Token.decimal    outerLexer
oStringLiteral = Text.Parsec.Token.stringLiteral outerLexer

typeAtom :: Parsec String InnerState Type
typeAtom  =   iParens typeExpr
          <|> do {
                st <- getState;
                i <- iIdentifier;
                return (typeIdent (stSimpleTypes st) i);
              }

typeExpr :: Parsec String InnerState Type
typeExpr  =   try (do {
                a <- typeAtom;
                iReservedOp "=>";
                r <- typeExpr;
                return (typeArrow a r);
              })
          <|> typeAtom

typeParser :: Parsec String InnerState Type
typeParser  = do {
                iWhiteSpace;
                result <- typeExpr;
                eof;
                return result;
              }

exprParser :: Parsec String InnerState Expr
exprParser  = do {
                st <- getState;
                iWhiteSpace;
                result <- let {
                  atom = iParens expr <|> do {
                    st <- getState;
                    i <- iIdentifier;
                    return (exprIdent (stConsts st) i);
                  } <|> choice (map (\(i, s) -> do { iReservedOp s; return (exprIdent (stConsts st) i); }) (Data.Set.toList $ stConstMixfixes st));

                  f (Prefix ms) = map (\(i, s) -> Text.Parsec.Expr.Prefix (do { iReservedOp s; return (\e   -> exprApply (exprIdent (stConsts st) i) e);               })                            ) ms;
                  f (Infix  ms) = map (\(i, s) -> Text.Parsec.Expr.Infix  (do { iReservedOp s; return (\a b -> exprApply (exprApply (exprIdent (stConsts st) i) a) b); }) Text.Parsec.Expr.AssocNone ) ms;
                  f (Infixl ms) = map (\(i, s) -> Text.Parsec.Expr.Infix  (do { iReservedOp s; return (\a b -> exprApply (exprApply (exprIdent (stConsts st) i) a) b); }) Text.Parsec.Expr.AssocLeft ) ms;
                  f (Infixr ms) = map (\(i, s) -> Text.Parsec.Expr.Infix  (do { iReservedOp s; return (\a b -> exprApply (exprApply (exprIdent (stConsts st) i) a) b); }) Text.Parsec.Expr.AssocRight) ms;

                  table = [[Text.Parsec.Expr.Infix (return exprApply) Text.Parsec.Expr.AssocLeft]] ++
                    (map f $ reverse $ map snd $ Data.Map.toList $ stMixfixes st) ++
                    [[Text.Parsec.Expr.Postfix (do { iReservedOp "::"; t <- typeExpr; return (\e -> exprTyped e t); })]];

                  expr = Text.Parsec.Expr.buildExpressionParser table atom <|> do {
                    iReservedOp "%";
                    x <- iIdentifier;
                    iReservedOp ".";
                    e <- expr;
                    return (exprLambda x e);
                  } <|> choice (map (\(i, s) -> do {
                    iReservedOp s;
                    x <- iIdentifier;
                    iReservedOp ".";
                    e <- expr;
                    return (exprApply (exprIdent (stConsts st) i) (exprLambda x e));
                  }) (Data.Set.toList $ stBinders st));
                } in expr;
                eof;
                return result;
              }

-- TODO: нормальные сообщения об ошибках
myRunParser :: Parsec String st ret -> st -> String -> String -> ret
myRunParser parser st document lang =
  case runParser parser st "-" document of {
    Left err -> error $ lang ++ ": " ++ Text.Parsec.Error.showErrorMessages "or" "???" "expected" "found" "nothing" (Text.Parsec.Error.errorMessages err);
    Right result -> result;
  }

oType :: Parsec String OuterState Type
oType = do {
          st <- getState;
          s <- oIdentifier <|> oStringLiteral;
          return $ myRunParser typeParser (stInner st) s ("type \"" ++ s ++ "\"");
        }

oExpr :: Parsec String OuterState Expr
oExpr = do {
          st <- getState;
          s <- (oIdentifier <|> oStringLiteral);
          return $ myRunParser exprParser (stInner st) s ("expr \"" ++ s ++ "\"");
        }

oInt :: Parsec String OuterState Int
oInt  = do {
          i <- oDecimal;
          -- TODO: как-то проверять overflow?
          return (fromInteger i);
        }

-- TODO: не проверяем даже, что такой ID есть. И полно таких же косяков
constDecl :: Parsec String OuterState ()
constDecl = do {
              st <- getState;
              i <- oIdentifier;
              oReservedOp "::";
              t <- oType;
              if Data.Map.member i (stConsts $ stInner st)
                then error ("const \"" ++ i ++ "\" redeclared")
                else setState st { stInner = (stInner st) { stConsts = Data.Map.insert i t $ stConsts $ stInner st } };
              option () $ do {
                st <- getState;
                oReservedOp "(";
                do {
                  s <- oStringLiteral;
                  setState st { stInner = (stInner st) { stConstMixfixes = Data.Set.insert (i, s) $ stConstMixfixes $ stInner st } };
                  return ();
                } <|> do {
                  oReserved "prefix";
                  s <- oStringLiteral;
                  p <- oInt;
                  case Data.Map.lookup p (stMixfixes $ stInner st) of {
                    Nothing          -> setState st { stInner = (stInner st) { stMixfixes = Data.Map.insert p (Prefix [(i, s)])    $ stMixfixes $ stInner st } };
                    Just (Prefix ms) -> setState st { stInner = (stInner st) { stMixfixes = Data.Map.insert p (Prefix ((i, s):ms)) $ stMixfixes $ stInner st } };
                    _ -> error $ i ++ ": different fixity for the same priority";
                  };
                  return ();
                } <|> do {
                  oReserved "infix";
                  s <- oStringLiteral;
                  p <- oInt;
                  case Data.Map.lookup p (stMixfixes $ stInner st) of {
                    Nothing          -> setState st { stInner = (stInner st) { stMixfixes = Data.Map.insert p (Infix  [(i, s)])    $ stMixfixes $ stInner st } };
                    Just (Infix  ms) -> setState st { stInner = (stInner st) { stMixfixes = Data.Map.insert p (Infix  ((i, s):ms)) $ stMixfixes $ stInner st } };
                    _ -> error $ i ++ ": different fixity for the same priority";
                  };
                  return ();
                } <|> do {
                  oReserved "infixl";
                  s <- oStringLiteral;
                  p <- oInt;
                  case Data.Map.lookup p (stMixfixes $ stInner st) of {
                    Nothing          -> setState st { stInner = (stInner st) { stMixfixes = Data.Map.insert p (Infixl [(i, s)])    $ stMixfixes $ stInner st } };
                    Just (Infixl ms) -> setState st { stInner = (stInner st) { stMixfixes = Data.Map.insert p (Infixl ((i, s):ms)) $ stMixfixes $ stInner st } };
                    _ -> error $ i ++ ": different fixity for the same priority";
                  };
                  return ();
                } <|> do {
                  oReserved "infixr";
                  s <- oStringLiteral;
                  p <- oInt;
                  case Data.Map.lookup p (stMixfixes $ stInner st) of {
                    Nothing          -> setState st { stInner = (stInner st) { stMixfixes = Data.Map.insert p (Infixr [(i, s)])    $ stMixfixes $ stInner st } };
                    Just (Infixr ms) -> setState st { stInner = (stInner st) { stMixfixes = Data.Map.insert p (Infixr ((i, s):ms)) $ stMixfixes $ stInner st } };
                    _ -> error $ i ++ ": different fixity for the same priority";
                  };
                  return ();
                } <|> do {
                  oReserved "binder";
                  s <- oStringLiteral;
                  setState st { stInner = (stInner st) { stBinders = Data.Set.insert (i, s) $ stBinders $ stInner st } };
                  return ();
                };
                oReservedOp ")";
                return ();
              };
              return ();
            }

axiom :: Parsec String OuterState ()
axiom = do {
          st <- getState;
          i <- oIdentifier;
          oReservedOp ":";
          (ps, c) <- (
                try (do {
                  p <- oExpr;
                  oReservedOp "==>";
                  c <- oExpr;
                  return ([p], c);
                })
            <|> do {
                  c <- oExpr;
                  return ([], c);
                }
            <|> do {
                  oReservedOp "[|";
                  ps <- sepBy1 oExpr (oReservedOp ";");
                  oReservedOp "|]";
                  oReservedOp "==>";
                  c <- oExpr;
                  return (ps, c);
                }
          );
          setState st { stTheorems = Data.Map.insert i (proofGeneralize (proofRule i ps c)) $ stTheorems st };
          return ();
        }
