import qualified Data.Set
import qualified Data.Map
import qualified Control.Exception
import qualified Control.DeepSeq
import Text.Parsec

import Kernel
import Parser
import Etc

decl :: Parsec String OuterState ()
decl  =   do {
            oReserved "consts";
            many1 constDecl;
            return ();
          }
      <|> do {
            oReserved "axioms";
            many1 axiom;
            return ();
          }
      <|> do {
            st <- getState;
            oReserved "lemma" <|> oReserved "theorem";
            theoremName <- oIdentifier;
            oReservedOp ":";
            (ps, c) <- (
                  do {
                    ps <- many (do {
                      oReserved "assumes";
                      i <- oIdentifier;
                      oReservedOp ":";
                      p <- oExpr;
                      return (i, p);
                    });
                    oReserved "shows";
                    c <- oExpr;
                    return (ps, c);
                  }
              <|> do {
                    c <- oExpr;
                    return ([], c);
                  }
            );
            c2:ps2 <- return (assumesShows (c:(map snd ps)));
            m <- return (Data.Map.fromList (zip (map fst ps) (map proofAssume ps2)));
            r <- proof;
            setState st { stTheorems = Data.Map.insert theoremName (proofGeneralize (proofAssms ps2 (checkReason st m c2 r))) (stTheorems st) };
            return ();
          }
      <|> etcDecl

documentParser :: Parsec String OuterState OuterState
documentParser  = do {
                    oWhiteSpace;
                    many decl;
                    eof;
                    st <- getState;
                    return st; -- Важно вернуть не только теоремы, но и, скажем, константы, чтобы убедиться, что они корректные
                  }

main =
  do {
    document <- getContents;

    Control.Exception.evaluate $ Control.DeepSeq.rnf $ myRunParser documentParser (OuterState (InnerState (Data.Set.singleton "prop") Data.Map.empty Data.Set.empty Data.Set.empty Data.Map.empty) Data.Map.empty emptyEtcState) document "outer language";

    return ();
  }
