#!/bin/sh

set -e

if [ $# != 1 ]; then
	echo "Usage: ${0##*/} PROGRAM" >&2
	exit 1
fi

PROGRAM="$1"

t(){
	printf "%s...\\n" "$1"
	tee "$1" | time -p "$PROGRAM"
	printf "Done\\n"
}

# "4 :: 0"          3
# "% 1000 :: 0. 3"  3
# "!! 1000 :: 0. 0" 0
# "2 ==> 1"         1

cat << EOF | t positive-implicational.pa
sorts kind type prop polymorphic

axioms
	type :: kind
	prop :: type

rules
	(prop, prop, prop)
	(type, prop, prop)
	(kind, prop, prop)
	(type, type, type)
	(kind, type, polymorphic)

{
	fix A :: prop
	fix B :: prop
	{
		fix a :: A
		fix B
		abbr a
	}
	abbr "this :: A ==> (B ==> A)"
}
abbr A1 = this

{
	fix A :: prop
	fix B :: prop
	fix C :: prop
	{
		fix abc :: "A ==> (B ==> C)"
		fix ab :: "A ==> B"
		fix a :: A
		abbr "abc a (ab a) :: C"
	}
	abbr "this :: (A ==> (B ==> C)) ==> ((A ==> B) ==> (A ==> C))"
}
abbr A2 = this

{
	fix A :: prop
	{
		fix A
	}
	abbr "this :: A ==> A"
}
abbr a_impl_a = this

{
	fix A :: prop
	fix B :: prop
	fix C :: prop
	{
		fix ab :: "A ==> B"
		fix bc :: "B ==> C"
		fix A
		abbr "ab this :: B"
		abbr "bc this :: C"
	}
	abbr "this :: (A ==> B) ==> ((B ==> C) ==> (A ==> C))"
}
abbr syllogism = this

{
	fix A :: prop
	fix B :: prop
	fix C :: prop
	{
		fix abc :: "A ==> (B ==> C)"
		fix b :: B
		fix A
		abbr "abc this :: B ==> C"
		abbr "this b :: C"
	}
	abbr "this :: (A ==> (B ==> C)) ==> (B ==> (A ==> C))"
}
abbr swap_premises = this

abbr reverse_syllogism = "% A :: prop. % B :: prop. % C :: prop. (swap_premises (A ==> B) (B ==> C) (A ==> C) (syllogism A B C) :: (B ==> C) ==> ((A ==> B) ==> (A ==> C)))"
EOF

cat positive-implicational.pa - << EOF | t positive.pa
fix conj :: "prop ==> (prop ==> prop)" (infixl "&" 35)
fix disj :: "prop ==> (prop ==> prop)" (infixl "|" 30)

fix A3 :: "!! A :: prop. !! B :: prop. A & B ==> A"
fix A4 :: "!! A :: prop. !! B :: prop. A & B ==> B"
fix A5 :: "!! A :: prop. !! B :: prop. A ==> (B ==> A & B)"
fix A6 :: "!! A :: prop. !! B :: prop. A ==> A | B"
fix A7 :: "!! A :: prop. !! B :: prop. B ==> A | B"
fix A8 :: "!! A :: prop. !! B :: prop. !! C :: prop. (A ==> C) ==> ((B ==> C) ==> (A | B ==> C))"
EOF

cat positive.pa - << EOF | t minimal.pa
fix Not :: "prop ==> prop" (prefix "~" 40)

fix A9 :: "!! A :: prop. !! B :: prop. (A ==> B) ==> ((A ==> ~B) ==> ~A)"

{
	fix A :: prop
	fix B :: prop
	{
		fix "A ==> B"
		abbr "A9 A B this :: (A ==> ~B) ==> ~A"
		abbr "syllogism (~B) (A ==> ~B) (~A) (A1 (~B) A) this :: ~B ==> ~A"
	}
	abbr "this :: (A ==> B) ==> (~B ==> ~A)"
}
abbr contraposition = this

{
	fix A :: prop
	{
		fix "A"
		abbr "A1 A (~A) this :: ~A ==> A"
		abbr "A9 (~A) A this (a_impl_a (~A)) :: ~~A"
	}
	abbr "this :: A ==> ~~A"
}
abbr intro_2_neg = this

{
	fix A :: prop
	abbr a = "A9 A (~~A) (intro_2_neg A) :: (A ==> ~~~A) ==> ~A"
	{
		fix "~~~A"
		abbr "A1 (~~~A) A this :: A ==> ~~~A"
		abbr "a this :: ~A"
	}
	abbr "this :: ~~~A ==> ~A"
}
abbr elim_3_neg = this

{
	fix A :: prop
	fix B :: prop
	{
		fix "A"
		abbr ba = "A1 A B this :: B ==> A"
		{
			fix "~A"
			abbr "A1 (~A) B this :: B ==> ~A"
			abbr "A9 B A ba this :: ~B"
		}
	}
	abbr "this :: A ==> (~A ==> ~B)"
}
abbr ex_falso_not = this
EOF

# Если к positive добавить константу False и никаких аксиом, получится логика, эквивалентная minimal, я не буду это доказывать

# Будем называть vml2010 конспекты лекций 1 - 8 курса "Введение в математическую логику" Л. Д. Беклемишева (http://lpcs.math.msu.su/vml2010 ). Также будем называть vml2010 формулировку классического исчисления высказываний, введённую в этих конспектах (10 аксиом A1 - A10 и MP)

cat minimal.pa - << EOF | t classical.pa
(* vml2010 *)

fix A10 :: "!! A :: prop. ~~A ==> A"

(* Это правило формализует "классическое рассуждение", противопоставленное интуиционистскому. Это формализация рассуждения "Если sqrt(2)^sqrt(2) рационально, то существуют рациональные числа вида a^b, где a и b иррациональны. Если иррационально - то тоже существуют. Значит, они существуют", которое обычно приводят в пример "неправильного" рассуждения интуционисты *)

{
	fix A :: prop
	fix B :: prop
	{
		fix "A ==> B"
		abbr "contraposition A B this :: ~B ==> ~A"
		abbr a = "A9 (~B) (~A) this :: (~B ==> ~~A) ==> ~~B"
		{
			fix "~A ==> B"
			abbr "contraposition (~A) B this :: ~B ==> ~~A"
			abbr "a this :: ~~B"
			abbr "A10 B this :: B"
		}
	}
	abbr "this :: (A ==> B) ==> ((~A ==> B) ==> B)"
}
abbr classical_reasoning = this

{
	fix A :: prop
	fix B :: prop
	abbr a = "A1 A (~B) :: A ==> (~B ==> A)"
	abbr "A9 (~B) A :: (~B ==> A) ==> ((~B ==> ~A) ==> ~~B)"
	abbr "syllogism A (~B ==> A) ((~B ==> ~A) ==> ~~B) a this :: A ==> ((~B ==> ~A) ==> ~~B)"
	abbr b = "swap_premises A (~B ==> ~A) (~~B) this :: (~B ==> ~A) ==> (A ==> ~~B)"
	abbr "A1 (~A) (~B) :: ~A ==> (~B ==> ~A)"
	abbr c = "syllogism (~A) (~B ==> ~A) (A ==> ~~B) this b :: ~A ==> (A ==> ~~B)"
	abbr "A10 B :: ~~B ==> B"
	abbr "reverse_syllogism A (~~B) B this :: (A ==> ~~B) ==> (A ==> B)"
	abbr "syllogism (~A) (A ==> ~~B) (A ==> B) c this :: ~A ==> (A ==> B)"
	abbr "swap_premises (~A) A B this :: A ==> (~A ==> B)"
}
abbr ex_falso = this

{
	fix A :: prop
	fix B :: prop
	abbr "A8 A B (~A ==> B) (ex_falso A B) (A1 B (~A)) :: A | B ==> (~A ==> B)"
}
abbr plan_b = this
EOF

cat classical.pa - << EOF | t false.pa
fix False :: prop

fix not_false :: "~False"

{
	fix A :: prop
	{
		fix "A ==> False"
		abbr "A1 (~False) A not_false :: A ==> ~False"
		abbr "A9 A False assm this :: ~A"
	}
	abbr "this :: (A ==> False) ==> ~A"
}
abbr contradiction2 = this
EOF

cat false.pa - << EOF | t fol.pa
fix i :: type

fix Ex :: "(i ==> prop) ==> prop" (binder "?" 10)

fix exI :: "!! A :: (i ==> prop). !! t :: i. A t ==> Ex A"
fix obtain :: "!! P :: (i ==> prop). !! Q :: prop. (!! x :: i. P x ==> Q) ==> (Ex P ==> Q)"
EOF

cat fol.pa - << EOF | t eq.pa
fix eq :: "i ==> (i ==> prop)" (infix "=" 50)

fix refl :: "!! a :: i. a = a"
fix subst :: "!! P :: i ==> prop. !! s :: i. !! t :: i. s = t ==> (P s ==> P t)"

{
	fix a :: i
	fix b :: i
	{
		fix "a = b"
		abbr "subst (%x :: i. x = a) a b :: a = b ==> (a = a ==> b = a)"
		abbr "this assm (refl a) :: b = a"
	}
	abbr "this :: a = b ==> b = a"
}
abbr eq_commute = this

{
	fix r :: i
	fix s :: i
	fix t :: i
	abbr "subst (%x :: i. x = t) s r :: s = r ==> (s = t ==> r = t)"
	abbr "syllogism (r = s) (s = r) (s = t ==> r = t) (eq_commute r s) this :: r = s ==> (s = t ==> r = t)"
}
abbr trans = this

{
	fix f :: "i ==> i"
	fix x :: i
	fix y :: i
	{
		fix "x = y"
		abbr a = "eq_commute x y this :: y = x"
		abbr "subst (%z :: i. f z = f y) y x :: y = x ==> (f y = f y ==> f x = f y)"
		abbr "this a (refl (f y)) :: f x = f y"
	}
	abbr "this :: x = y ==> f x = f y"
}
abbr arg_cong = this
EOF

cat eq.pa - << EOF | t pa.pa
fix zero :: i ("0")
fix succ :: "i ==> i"

fix PA1 :: "!! a :: i. ~ succ a = 0"
fix PA2 :: "!! a :: i. a = 0 | (? b :: i. a = succ b)"
fix PA3 :: "!! a :: i. !! b :: i. succ a = succ b ==> a = b"
fix PA4 :: "!! P :: (i ==> prop). !! a :: i. P 0 ==> ((!! x :: i. P x ==> P (succ x)) ==> P a)"

fix le :: "i ==> (i ==> prop)" (infix "<=" 50)

fix LE1 :: "!! a :: i. a <= 0 ==> a = 0"
fix LE2 :: "!! a :: i. a = 0 ==> a <= 0"
fix LE3 :: "!! a :: i. !! b :: i. a <= succ b ==> a = succ b | a <= b"
fix LE4 :: "!! a :: i. !! b :: i. a = succ b ==> a <= succ b"
fix LE5 :: "!! a :: i. !! b :: i. a <= b ==> a <= succ b"

fix less :: "i ==> (i ==> prop)" (infix "<" 50)

fix L1 :: "!! a :: i. !! b :: i. a < b ==> a <= b"
fix L2 :: "!! a :: i. !! b :: i. a < b ==> ~(a = b)"
fix L3 :: "!! a :: i. !! b :: i. a <= b ==> (~(a = b) ==> a < b)"

{
	fix a :: i
	{
		fix "a < 0"
		abbr "L1 a 0 this :: a <= 0"
		abbr "LE1 a this :: a = 0"
		abbr "ex_falso (a = 0) False this (L2 a 0 assm) :: False"
	}
	abbr "contradiction2 (a < 0) this :: ~ a < 0"
}
abbr less_zero = this

{
	fix a :: i
	fix b :: i
	{
		fix "a < succ b"
		abbr "L1 a (succ b) assm :: a <= succ b"
		abbr c = "LE3 a b this :: a = succ b | a <= b"
		abbr "L2 a (succ b) assm :: ~(a = succ b)"
		abbr "plan_b (a = succ b) (a <= b) c this :: a <= b"
	}
	abbr "this :: a < succ b ==> a <= b"
}
abbr less_succ = this

{
	fix a :: i
	fix b :: i
	{
		fix aleb :: "a <= b"
		abbr switch = "PA2 b :: b = 0 | (? x :: i. b = succ x)"
		{
			fix "b = 0"
			abbr "subst (%x :: i. a <= x) b 0 this aleb :: a <= 0"
			abbr "LE1 a this :: a = 0"
			abbr "trans a 0 b this (eq_commute b 0 assm) :: a = b"
			abbr "A7 (a < b) (a = b) this :: a < b | a = b"
		}
		abbr casea = this
		{
			fix x :: i
			fix obtaining :: "b = succ x"
			abbr "subst (%x :: i. a <= x) b (succ x) this aleb :: a <= succ x"
			abbr or = "LE3 a x this :: a = succ x | a <= x"
			{
				fix "a = succ x"
				abbr "trans a (succ x) b this (eq_commute b (succ x) obtaining) :: a = b"
				abbr "A7 (a < b) (a = b) this :: a < b | a = b"
			}
			abbr casea2 = this
			{
				fix "~ a = succ x"
				abbr "plan_b (a = succ x) (a <= x) or this :: a <= x"
				abbr "LE5 a x this :: a <= succ x"
				abbr "L3 a (succ x) this assm :: a < succ x"
				abbr "subst (%x :: i. a < x) (succ x) b (eq_commute b (succ x) obtaining) this :: a < b"
				abbr "A6 (a < b) (a = b) this :: a < b | a = b"
			}
			abbr caseb2 = this
			abbr "classical_reasoning (a = succ x) (a < b | a = b) casea2 caseb2 :: a < b | a = b"
		}
		abbr caseb = "obtain (%x :: i. b = succ x) (a < b | a = b) this :: (? x :: i. b = succ x) ==> a < b | a = b"
		abbr "A8 (b = 0) (? x :: i. b = succ x) (a < b | a = b) casea caseb switch :: a < b | a = b"
	}
	abbr "this :: a <= b ==> a < b | a = b"
}
abbr le = this

{
	fix P :: "i ==> prop"
	fix a :: i
	{
		fix assm1 :: "!! x :: i. (!! y :: i. y < x ==> P y) ==> P x"
		{
			fix t :: i
			abbr "swap_premises (t < 0) (~ t < 0) (P t) (ex_falso (t < 0) (P t)) (less_zero t) :: t < 0 ==> P t"
		}
		abbr base = "this :: !! t :: i. t < 0 ==> P t"
			{
				fix z :: i
				abbr a1 = "assm1 z :: (!! y :: i. y < z ==> P y) ==> P z"
				{
					fix assm2 :: "!! t :: i. t < z ==> P t"
					abbr pz = "a1 this :: P z"
					{
						fix t :: "i"
						abbr casea = "assm2 t :: t < z ==> P t"
						{
							fix "t < succ z"
							abbr "less_succ t z this :: t <= z"
							abbr switch = "le t z this :: t < z | t = z"
							{
								fix "t = z"
								abbr "subst P z t (eq_commute t z this) pz :: P t"
							}
							abbr caseb = this
							abbr "A8 (t < z) (t = z) (P t) casea caseb switch :: P t"
						}
						abbr "this :: t < succ z ==> P t"
					}
					abbr "this :: !! t :: i. t < succ z ==> P t"
				}
				abbr "this :: (!! t :: i. t < z ==> P t) ==> (!! t :: i. t < succ z ==> P t)"
			}
		abbr "PA4 (%a :: i. (!! t :: i. t < a ==> P t)) a base this :: !! t :: i. t < a ==> P t"
		abbr "assm1 a this :: P a"
	}
	abbr "this :: (!! x :: i. (!! y :: i. y < x ==> P y) ==> P x) ==> P a"
}
abbr full_induction = this
EOF
